import ast
import typing
from typing import Optional, Union


class AST(ast.AST):
    "Scenic AST base class"

    _attributes = ("lineno", "col_offset", "end_lineno", "end_col_offset")

    def __init__(self, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)


# special statements


class TryInterrupt(AST):
    """Scenic AST node that represents try-interrupt statements"""

    __match_args__ = (
        "body",
        "interrupt_when_handlers",
        "except_handlers",
        "orelse",
        "finalbody",
    )

    def __init__(
        self,
        body: typing.List[ast.stmt],
        interrupt_when_handlers: typing.List["InterruptWhenHandler"],
        except_handlers: typing.List[ast.ExceptHandler],
        orelse: typing.List[ast.stmt],
        finalbody: typing.List[ast.AST],
        *args: any,
        **kwargs: any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.body = body
        self.interrupt_when_handlers = interrupt_when_handlers
        self.except_handlers = except_handlers
        self.orelse = orelse
        self.finalbody = finalbody
        self._fields = [
            "body",
            "interrupt_when_handlers",
            "except_handlers",
            "orelse",
            "finalbody",
        ]
        self._attributes = []


class InterruptWhenHandler(AST):
    __match_args__ = ("cond", "body")

    def __init__(
        self, cond: ast.AST, body: typing.List[ast.AST], *args: any, **kwargs: any
    ) -> None:
        super().__init__(*args, **kwargs)
        self.cond = cond
        self.body = body
        self._fields = ["cond", "body"]


class TrackedAssign(AST):
    __match_args__ = (
        "target",
        "value",
    )

    def __init__(
        self, target: Union["Ego", "Workspace"], value: ast.AST, *args: any, **kwargs: any
    ) -> None:
        super().__init__(*args, **kwargs)
        self.target = target
        self.value = value
        self._fields = ["target", "value"]


class Ego(AST):
    "`ego` tracked assign target"
    functionName = "ego"


class Workspace(AST):
    ":term:`workspace` tracked assign target"
    functionName = "workspace"


class InitialScenario(AST):
    pass


class PropertyDef(AST):
    __match_args__ = ("property", "attributes", "value")

    def __init__(
        self,
        property: str,
        attributes: typing.List[Union["Additive", "Dynamic", "Final"]],
        value=ast.AST,
        *args: any,
        **kwargs: any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.property = property
        self.attributes = attributes
        self.value = value
        self._fields = ["property", "attributes", "value"]


class Additive(AST):
    keyword = "additive"

    def __init__(self, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)


class Dynamic(AST):
    keyword = "dynamic"

    def __init__(self, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)


class Final(AST):
    keyword = "final"

    def __init__(self, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)


# behavior / monitor


class BehaviorDef(AST):
    __match_args__ = ("name", "args", "docstring", "header", "body")

    def __init__(
        self,
        name: str,
        args: ast.arguments,
        docstring: Optional[str],
        header: typing.List[Union["Precondition", "Invariant"]],
        body: typing.List[any],
        *_args: any,
        **kwargs: any,
    ) -> None:
        super().__init__(*_args, **kwargs)
        self.name = name
        self.args = args
        self.docstring = docstring
        self.header = header
        self.body = body
        self._fields = ["name", "args", "docstring", "header", "body"]


class MonitorDef(AST):
    __match_args__ = ("name", "args", "docstring", "body")

    def __init__(
        self,
        name: str,
        args: ast.arguments,
        docstring: Optional[str],
        body: typing.List[ast.AST],
        *_args: any,
        **kwargs: any,
    ) -> None:
        super().__init__(*_args, **kwargs)
        self.name = name
        self.args = args
        self.docstring = docstring
        self.body = body
        self._fields = ["name", "args", "docstring", "body"]


class Precondition(AST):
    __match_args__ = ("value",)

    def __init__(self, value: ast.AST, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)
        self.value = value
        self._fields = ["value"]


class Invariant(AST):
    __match_args__ = ("value",)

    def __init__(self, value: ast.AST, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)
        self.value = value
        self._fields = ["value"]


# modular scenarios


class ScenarioDef(AST):
    __match_args__ = ("name", "args", "docstring", "header", "setup", "compose")

    def __init__(
        self,
        name: str,
        args: ast.arguments,
        docstring: Optional[str],
        header: Optional[typing.List[Union[Precondition, Invariant]]],
        setup: typing.List[ast.AST],
        compose: typing.List[ast.AST],
        *_args: any,
        **kwargs: any,
    ) -> None:
        super().__init__(*_args, **kwargs)
        self.name = name
        self.args = args
        self.docstring = docstring
        self.header = header
        self.setup = setup
        self.compose = compose
        self._fields = ["name", "args", "docstring", "header", "setup", "compose"]


# simple statements


class Model(AST):
    __match_args__ = ("name",)

    def __init__(self, name: str, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)
        self.name = name
        self._fields = ["name"]


class Param(AST):
    ":keyword:`param` statements"

    __match_args__ = ("elts",)

    def __init__(self, elts: typing.List["parameter"], *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)
        self.elts = elts
        self._fields = ["elts"]


class parameter(AST):
    "represents a parameter that is defined with `param` statements"
    __match_args__ = ("identifier", "value")

    def __init__(
        self, identifier: str, value: ast.AST, *args: any, **kwargs: any
    ) -> None:
        super().__init__(*args, **kwargs)
        self.identifier = identifier
        self.value = value
        self._fields = ["identifier", "value"]


class Require(AST):
    __match_args__ = ("cond", "prob", "name")

    def __init__(
        self,
        cond: ast.AST,
        prob: Optional[float] = None,
        name: Optional[str] = None,
        *args: any,
        **kwargs: any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.cond = cond
        self.prob = prob
        self.name = name
        self._fields = ["cond", "prob", "name"]


class RequireMonitor(AST):
    __match_args__ = ("monitor", "name")

    def __init__(
        self, monitor: ast.AST, name: Optional[str] = None, *args: any, **kwargs: any
    ) -> None:
        super().__init__(*args, **kwargs)
        self.monitor = monitor
        self.name = name
        self._fields = ["monitor", "name"]


class Always(AST):
    __match_args__ = ("value",)

    def __init__(self, value: ast.AST, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.value = value
        self._fields = ["value"]

    def __reduce__(self):
        return (
            type(self),
            (self.value,),
            {
                "lineno": self.lineno,
                "end_lineno": self.end_lineno,
                "col_offset": self.col_offset,
                "end_col_offset": self.end_col_offset,
            },
        )


class Eventually(AST):
    __match_args__ = ("value",)

    def __init__(self, value: ast.AST, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.value = value
        self._fields = ["value"]

    def __reduce__(self):
        return (
            type(self),
            (self.value,),
            {
                "lineno": self.lineno,
                "end_lineno": self.end_lineno,
                "col_offset": self.col_offset,
                "end_col_offset": self.end_col_offset,
            },
        )


class Next(AST):
    __match_args__ = ("value",)

    def __init__(self, value: ast.AST, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.value = value
        self._fields = ["value"]

    def __reduce__(self):
        return (
            type(self),
            (self.value,),
            {
                "lineno": self.lineno,
                "end_lineno": self.end_lineno,
                "col_offset": self.col_offset,
                "end_col_offset": self.end_col_offset,
            },
        )


class Record(AST):
    __match_args__ = ("value", "name")

    def __init__(
        self, value: ast.AST, name: Optional[str] = None, *args: any, **kwargs: any
    ) -> None:
        super().__init__(*args, **kwargs)
        self.value = value
        self.name = name
        self._fields = ["value", "name"]


class RecordInitial(AST):
    __match_args__ = ("value", "name")

    def __init__(
        self, value: ast.AST, name: Optional[str] = None, *args: any, **kwargs: any
    ) -> None:
        super().__init__(*args, **kwargs)
        self.value = value
        self.name = name
        self._fields = ["value", "name"]


class RecordFinal(AST):
    __match_args__ = ("value", "name")

    def __init__(
        self, value: ast.AST, name: Optional[str] = None, *args: any, **kwargs: any
    ) -> None:
        super().__init__(*args, **kwargs)
        self.value = value
        self.name = name
        self._fields = ["value", "name"]


class Mutate(AST):
    __match_args__ = ("elts", "scale")

    def __init__(
        self,
        elts: typing.List[ast.Name],
        scale: Optional[ast.AST] = None,
        *args: any,
        **kwargs: any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.elts = elts
        self.scale = scale
        self._fields = ["elts", "scale"]


class Override(AST):
    __match_args__ = ("target", "specifiers")

    def __init__(
        self, target: ast.AST, specifiers: typing.List[ast.AST], *args: any, **kwargs: any
    ) -> None:
        super().__init__(*args, **kwargs)
        self.target = target
        self.specifiers = specifiers
        self._fields = ["target", "specifiers"]


class Abort(AST):
    pass


class Take(AST):
    __match_args__ = ("elts",)

    def __init__(self, elts: typing.List[ast.AST], *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)
        self.elts = elts
        self._fields = ["elts"]


class Wait(AST):
    pass


class Terminate(AST):
    pass


class TerminateSimulation(AST):
    pass


class TerminateSimulationWhen(AST):
    __match_args__ = (
        "cond",
        "name",
    )

    def __init__(
        self, cond: ast.AST, name: Optional[str] = None, *args: any, **kwargs: any
    ) -> None:
        super().__init__(*args, **kwargs)
        self.cond = cond
        self.name = name
        self._fields = ["cond", "name"]


class TerminateWhen(AST):
    __match_args__ = (
        "cond",
        "name",
    )

    def __init__(
        self, cond: ast.AST, name: Optional[str] = None, *args: any, **kwargs: any
    ) -> None:
        super().__init__(*args, **kwargs)
        self.cond = cond
        self.name = name
        self._fields = ["cond", "name"]


class TerminateAfter(AST):
    __match_args__ = ("duration",)

    def __init__(
        self, duration: Union["Seconds", "Steps"], *args: any, **kwargs: any
    ) -> None:
        super().__init__(*args, **kwargs)
        self.duration = duration
        self._fields = ["duration"]


class DoFor(AST):
    __match_args__ = ("elts", "duration")

    def __init__(
        self,
        elts: typing.List[ast.AST],
        duration: Union["Seconds", "Steps"],
        *args: any,
        **kwargs: any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.elts = elts
        self.duration = duration
        self._fields = ["elts", "duration"]


class Seconds(AST):
    __match_args__ = ("value",)
    unitStr = "seconds"

    def __init__(self, value: ast.AST, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)
        self.value = value
        self._fields = ["value"]


class Steps(AST):
    __match_args__ = ("value",)
    unitStr = "steps"

    def __init__(self, value: ast.AST, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)
        self.value = value
        self._fields = ["value"]


class DoUntil(AST):
    __match_args__ = ("elts", "cond")

    def __init__(
        self, elts: typing.List[ast.AST], cond: ast.AST, *args: any, **kwargs: any
    ) -> None:
        super().__init__(*args, **kwargs)
        self.elts = elts
        self.cond = cond
        self._fields = ["elts", "cond"]


class DoChoose(AST):
    __match_args__ = ("elts",)

    def __init__(self, elts: typing.List[ast.AST], *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)
        self.elts = elts
        self._fields = ["elts"]


class DoShuffle(AST):
    __match_args__ = ("elts",)

    def __init__(self, elts: typing.List[ast.AST], *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)
        self.elts = elts
        self._fields = ["elts"]


class Do(AST):
    __match_args__ = ("elts",)

    def __init__(self, elts: typing.List[ast.AST], *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)
        self.elts = elts
        self._fields = ["elts"]


class Simulator(AST):
    __match_args__ = ("value",)

    def __init__(self, value: ast.AST, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)
        self.value = value
        self._fields = ["value"]


# Instance Creation


class New(AST):
    __match_args__ = ("className", "specifiers")

    def __init__(
        self, className: str, specifiers: list, *args: any, **kwargs: any
    ) -> None:
        super().__init__(*args, **kwargs)
        self.className = className
        self.specifiers = specifiers if specifiers is not None else []
        self._fields = ["className", "specifiers"]


# Specifiers


class WithSpecifier(AST):
    __match_args__ = ("prop", "value")

    def __init__(self, prop: str, value: ast.AST, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)
        self.prop = prop
        self.value = value
        self._fields = ["prob", "value"]


class AtSpecifier(AST):
    __match_args__ = ("position",)

    def __init__(self, position: ast.AST, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)
        self.position = position
        self._fields = ["position"]


class OffsetBySpecifier(AST):
    __match_args__ = ("offset",)

    def __init__(self, offset: ast.AST, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)
        self.offset = offset
        self._fields = ["offset"]


class OffsetAlongSpecifier(AST):
    __match_args__ = ("direction", "offset")

    def __init__(
        self, direction: ast.AST, offset: ast.AST, *args: any, **kwargs: any
    ) -> None:
        super().__init__(*args, **kwargs)
        self.direction = direction
        self.offset = offset
        self._fields = ["direction", "offset"]


class DirectionOfSpecifier(AST):
    __match_args__ = ("direction", "position", "distance")

    def __init__(
        self,
        direction: Union["LeftOf", "RightOf", "AheadOf", "Behind"],
        position: ast.AST,
        distance: Optional[ast.AST],
        *args: any,
        **kwargs: any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.direction = direction
        self.position = position
        self.distance = distance
        self._fields = ["direction", "position", "distance"]


class LeftOf(AST):
    def __init__(self, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)


class RightOf(AST):
    def __init__(self, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)


class AheadOf(AST):
    def __init__(self, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)


class Behind(AST):
    def __init__(self, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)


class Above(AST):
    def __init__(self, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)


class Below(AST):
    def __init__(self, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)


class BeyondSpecifier(AST):
    __match_args__ = ("position", "offset", "base")

    def __init__(
        self,
        position: ast.AST,
        offset: ast.AST,
        base: Optional[ast.AST] = None,
        *args: any,
        **kwargs: any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.position = position
        self.offset = offset
        self.base = base
        self._fields = ["position", "offset", "base"]


class VisibleSpecifier(AST):
    __match_args__ = ("base",)

    def __init__(self, base: Optional[ast.AST] = None, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)
        self.base = base
        self._fields = ["base"]


class NotVisibleSpecifier(AST):
    __match_args__ = ("base",)

    def __init__(self, base: Optional[ast.AST] = None, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)
        self.base = base
        self._fields = ["base"]


class InSpecifier(AST):
    __match_args__ = ("region",)

    def __init__(self, region: ast.AST, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)
        self.region = region
        self._fields = ["region"]


class OnSpecifier(AST):
    __match_args__ = ("region",)

    def __init__(self, region: ast.AST, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)
        self.region = region
        self._fields = ["region"]


class ContainedInSpecifier(AST):
    __match_args__ = ("region",)

    def __init__(self, region: ast.AST, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)
        self.region = region
        self._fields = ["region"]


class FollowingSpecifier(AST):
    __match_args__ = ("field", "distance", "base")

    def __init__(
        self,
        field: ast.AST,
        distance: ast.AST,
        base: Optional[ast.AST] = None,
        *args: any,
        **kwargs: any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.field = field
        self.distance = distance
        self.base = base
        self._fields = ["field", "distance", "base"]


class FacingSpecifier(AST):
    __match_args__ = ("heading",)

    def __init__(self, heading: ast.AST, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)
        self.heading = heading
        self._fields = ["heading"]


class FacingTowardSpecifier(AST):
    __match_args__ = ("position",)

    def __init__(self, position: ast.AST, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)
        self.position = position
        self._fields = ["position"]


class FacingAwayFromSpecifier(AST):
    __match_args__ = ("position",)

    def __init__(self, position: ast.AST, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)
        self.position = position
        self._fields = ["position"]


class FacingDirectlyTowardSpecifier(AST):
    __match_args__ = ("position",)

    def __init__(self, position: ast.AST, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)
        self.position = position
        self._fields = ["position"]


class FacingDirectlyAwayFromSpecifier(AST):
    __match_args__ = ("position",)

    def __init__(self, position: ast.AST, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)
        self.position = position
        self._fields = ["position"]


class ApparentlyFacingSpecifier(AST):
    __match_args__ = ("heading", "base")

    def __init__(
        self, heading: ast.AST, base: Optional[ast.AST] = None, *args: any, **kwargs: any
    ) -> None:
        super().__init__(*args, **kwargs)
        self.heading = heading
        self.base = base
        self._fields = ["heading", "base"]


# Operators


class ImpliesOp(AST):
    __match_args__ = ("hypothesis", "conclusion")

    def __init__(
        self, hypothesis: ast.AST, conclusion: ast.AST, *args: any, **kwargs: any
    ) -> None:
        super().__init__(*args, **kwargs)
        self.hypothesis = hypothesis
        self.conclusion = conclusion
        self._fields = ["hypothesis", "conclusion"]

    def __reduce__(self):
        return (
            type(self),
            (self.hypothesis, self.conclusion),
            {
                "lineno": self.lineno,
                "end_lineno": self.end_lineno,
                "col_offset": self.col_offset,
                "end_col_offset": self.end_col_offset,
            },
        )


class UntilOp(AST):
    __match_args__ = ("left", "right")

    def __init__(self, left: ast.AST, right: ast.AST, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)
        self.left = left
        self.right = right
        self._fields = ["left", "right"]

    def __reduce__(self):
        return (
            type(self),
            (self.left, self.right),
            {
                "lineno": self.lineno,
                "end_lineno": self.end_lineno,
                "col_offset": self.col_offset,
                "end_col_offset": self.end_col_offset,
            },
        )


class RelativePositionOp(AST):
    __match_args__ = ("target", "base")

    def __init__(
        self, target: ast.AST, base: ast.AST = None, *args: any, **kwargs: any
    ) -> None:
        super().__init__(*args, **kwargs)
        self.target = target
        self.base = base
        self._fields = ["target", "base"]


class RelativeHeadingOp(AST):
    __match_args__ = ("target", "base")

    def __init__(
        self, target: ast.AST, base: ast.AST = None, *args: any, **kwargs: any
    ) -> None:
        super().__init__(*args, **kwargs)
        self.target = target
        self.base = base
        self._fields = ["target", "base"]


class ApparentHeadingOp(AST):
    __match_args__ = ("target", "base")

    def __init__(
        self, target: ast.AST, base: ast.AST = None, *args: any, **kwargs: any
    ) -> None:
        super().__init__(*args, **kwargs)
        self.target = target
        self.base = base
        self._fields = ["target", "base"]


class DistanceFromOp(AST):
    __match_args__ = ("target", "base")

    def __init__(
        self,
        # because `to` and `from` are symmetric, the first operand will be `target` and the second will be `base`
        target: ast.AST,
        base: Optional[ast.AST] = None,
        *args: any,
        **kwargs: any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.target = target
        self.base = base
        self._fields = ["target", "base"]


class DistancePastOp(AST):
    __match_args__ = ("target", "base")

    def __init__(
        self, target: ast.AST, base: Optional[ast.AST] = None, *args: any, **kwargs: any
    ) -> None:
        super().__init__(*args, **kwargs)
        self.target = target
        self.base = base
        self._fields = ["target", "base"]


class AngleFromOp(AST):
    __match_args__ = ("target", "base")

    def __init__(
        self,
        target: Optional[ast.AST] = None,
        base: Optional[ast.AST] = None,
        *args: any,
        **kwargs: any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.target = target
        self.base = base
        self._fields = ["target", "base"]


class AltitudeFromOp(AST):
    __match_args__ = ("target", "base")

    def __init__(
        self,
        target: Optional[ast.AST] = None,
        base: Optional[ast.AST] = None,
        *args: any,
        **kwargs: any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.target = target
        self.base = base
        self._fields = ["target", "base"]


class FollowOp(AST):
    __match_args__ = ("target", "base", "distance")

    def __init__(
        self, target: ast.AST, base: ast.AST, distance: ast.AST, *args: any, **kwargs: any
    ) -> None:
        super().__init__(*args, **kwargs)
        self.target = target
        self.base = base
        self.distance = distance
        self._fields = ["target", "base", "distance"]


class VisibleOp(AST):
    __match_args__ = ("region",)

    def __init__(self, region: ast.AST, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)
        self.region = region
        self._fields = ["region"]


class NotVisibleOp(AST):
    __match_args__ = ("region",)

    def __init__(self, region: ast.AST, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)
        self.region = region
        self._fields = ["region"]


class VisibleFromOp(AST):
    __match_args__ = ("region", "base")

    def __init__(self, region: ast.AST, base: ast.AST, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)
        self.region = region
        self.base = base
        self._fields = ["region", "base"]


class NotVisibleFromOp(AST):
    __match_args__ = ("region", "base")

    def __init__(self, region: ast.AST, base: ast.AST, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)
        self.region = region
        self.base = base
        self._fields = ["region", "base"]


class PositionOfOp(AST):
    __match_args__ = ("position", "target")

    def __init__(
        self,
        position: Union[
            "Front",
            "Back",
            "Left",
            "Right",
            "Top",
            "Bottom",
            "FrontLeft",
            "FrontRight",
            "BackLeft",
            "BackRight",
            "TopFrontLeft",
            "TopFrontRight",
            "TopBackLeft",
            "TopBackRight",
            "BottomFrontLeft",
            "BottomFrontRight",
            "BottomBackLeft",
            "BottomBackRight",
        ],
        target: ast.AST,
        *args: any,
        **kwargs: any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.position = position
        self.target = target
        self._fields = ["position", "target"]


class Front(AST):
    "Represents position of :scenic:`front of` operator"
    functionName = "Front"

    def __init__(self, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)


class Back(AST):
    "Represents position of :scenic:`back of` operator"
    functionName = "Back"

    def __init__(self, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)


class Left(AST):
    "Represents position of :scenic:`left of` operator"
    functionName = "Left"

    def __init__(self, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)


class Right(AST):
    "Represents position of :scenic:`right of` operator"
    functionName = "Right"

    def __init__(self, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)


class Top(AST):
    "Represents position of :scenic:`top of` operator"
    functionName = "Top"

    def __init__(self, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)


class Bottom(AST):
    "Represents position of :scenic:`bottom of` operator"
    functionName = "Bottom"

    def __init__(self, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)


class FrontLeft(AST):
    "Represents position of :scenic:`front left of` operator"
    functionName = "FrontLeft"

    def __init__(self, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)


class FrontRight(AST):
    "Represents position of :scenic:`front right of` operator"
    functionName = "FrontRight"

    def __init__(self, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)


class BackLeft(AST):
    "Represents position of :scenic:`back left of` operator"
    functionName = "BackLeft"

    def __init__(self, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)


class BackRight(AST):
    "Represents position of :scenic:`back right of` operator"
    functionName = "BackRight"

    def __init__(self, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)


class TopFrontLeft(AST):
    "Represents position of :scenic:`top front left of` operator"
    functionName = "TopFrontLeft"

    def __init__(self, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)


class TopFrontRight(AST):
    "Represents position of :scenic:`top front right of` operator"
    functionName = "TopFrontRight"

    def __init__(self, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)


class TopBackLeft(AST):
    "Represents position of :scenic:`top back left of` operator"
    functionName = "TopBackLeft"

    def __init__(self, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)


class TopBackRight(AST):
    "Represents position of :scenic:`top back right of` operator"
    functionName = "TopBackRight"

    def __init__(self, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)


class BottomFrontLeft(AST):
    "Represents position of :scenic:`bottom front left of` operator"
    functionName = "BottomFrontLeft"

    def __init__(self, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)


class BottomFrontRight(AST):
    "Represents position of :scenic:`bottom front right of` operator"
    functionName = "BottomFrontRight"

    def __init__(self, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)


class BottomBackLeft(AST):
    "Represents position of :scenic:`bottom back left of` operator"
    functionName = "BottomBackLeft"

    def __init__(self, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)


class BottomBackRight(AST):
    "Represents position of :scenic:`bottom back right of` operator"
    functionName = "BottomBackRight"

    def __init__(self, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)


class DegOp(AST):
    __match_args__ = ("operand",)

    def __init__(self, operand: ast.AST, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)
        self.operand = operand
        self._fields = ["operand"]


class VectorOp(AST):
    __match_args__ = ("left", "right")

    def __init__(self, left: ast.AST, right: ast.AST, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)
        self.left = left
        self.right = right
        self._fields = ["left", "right"]


class FieldAtOp(AST):
    __match_args__ = ("left", "right")

    def __init__(self, left: ast.AST, right: ast.AST, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)
        self.left = left
        self.right = right
        self._fields = ["left", "right"]


class RelativeToOp(AST):
    __match_args__ = ("left", "right")

    def __init__(self, left: ast.AST, right: ast.AST, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)
        self.left = left
        self.right = right
        self._fields = ["left", "right"]


class OffsetAlongOp(AST):
    __match_args__ = ("base", "direction", "offset")

    def __init__(
        self,
        base: ast.AST,
        direction: ast.AST,
        offset: ast.AST,
        *args: any,
        **kwargs: any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.base = base
        self.direction = direction
        self.offset = offset
        self._fields = ["base", "direction", "offset"]


class CanSeeOp(AST):
    __match_args__ = ("left", "right")

    def __init__(self, left: ast.AST, right: ast.AST, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)
        self.left = left
        self.right = right
        self._fields = ["left", "right"]


class IntersectsOp(AST):
    __match_args__ = ("left", "right")

    def __init__(self, left: ast.AST, right: ast.AST, *args: any, **kwargs: any) -> None:
        super().__init__(*args, **kwargs)
        self.left = left
        self.right = right
        self._fields = ["left", "right"]



"""Pygments lexer and style for Scenic.

These work with the `Pygments syntax highlighter <https://pygments.org/>`_.
The module actually defines several lexers used for the Scenic documentation;
the main `ScenicLexer` and its associated style `ScenicStyle` are exported by
:file:`pyproject.toml` as plugins to Pygments. This means that if you have the
``scenic`` package installed, the Pygments command-line tool and Python API
will automatically recognize Scenic files. For example, to highlight a Scenic
program as a self-contained HTML or LaTeX file:

.. code-block:: console

    $ pygmentize -f html -Ofull,style=scenic prog.scenic > out.html
    $ pygmentize -f latex -Ofull,style=scenic prog.scenic > out.tex

If highlighting multiple pieces of code, remove the ``full`` option to avoid
having the requisite CSS/preamble material duplicated in all your outputs; you
can run :command:`pygmentize -S scenic -f html` (or :command:`latex`) to
generate that material separately.
"""

import re

try:
    from pygments.lexers.python import PythonLexer
except ModuleNotFoundError as e:
    raise ImportError(
        "need the 'pygments' package to use the Scenic Pygments lexer"
    ) from e
from pygments.lexer import (
    RegexLexer,
    bygroups,
    combined,
    default,
    include,
    inherit,
    using,
    words,
)
from pygments.style import Style
from pygments.styles.default import DefaultStyle
from pygments.token import *
import pygments.unistring as uni


class BetterPythonLexer(PythonLexer):
    """Python lexer with better highlighting of function calls, parameters, etc.

    OK, 'better' is a matter of opinion; but it provides more informative tokens.
    These tokens will not cause errors under any Pygments style, but require the
    style to be aware of them in order to actually get better highlighting: use
    the `ScenicStyle` below for best results.

    Adapted from the PythonLexer and the MagicPython grammar by MagicStack Inc.,
    available at https://github.com/MagicStack/MagicPython.
    """

    name = "Python (a la MagicPython)"
    aliases = ["better-python"]

    uni_name = PythonLexer.uni_name

    tokens = {
        "root": [
            # This is largely copied from PythonLexer: the differences are noted below
            (r"\n", Whitespace),
            (
                r'^(\s*)([rRuUbB]{,2})("""(?:.|\n)*?""")',
                bygroups(Whitespace, String.Affix, String.Doc),
            ),
            (
                r"^(\s*)([rRuUbB]{,2})('''(?:.|\n)*?''')",
                bygroups(Whitespace, String.Affix, String.Doc),
            ),
            (r"\A#!.+$", Comment.Hashbang),
            (r"#.*$", Comment.Single),
            (r"\\\n", Punctuation),  # we count line continuations as punctuation
            (r"\\", Punctuation),  # ditto
            (r"\b(lambda)\b", Keyword, "lambda"),  # parse parameters of lambdas
            include("keywords"),
            include("soft-keywords"),
            (r"(def)((?:\s|\\\s)+)", bygroups(Keyword, Punctuation), "funcname"),
            (r"(class)((?:\s|\\\s)+)", bygroups(Keyword, Punctuation), "classname"),
            (
                r"(from)((?:\s|\\\s)+)",
                bygroups(Keyword.Namespace, Punctuation),
                "fromimport",
            ),
            (
                r"(import)((?:\s|\\\s)+)",
                bygroups(Keyword.Namespace, Punctuation),
                "import",
            ),
            include("extra-statements"),  # hook for subclasses
            # Parse expressions inside decorators
            (
                r"^(\s*)(\@)(\s*)",
                bygroups(Whitespace, Name.Decorator, Whitespace),
                "decorator",
            ),
            # Finally, we can have ordinary expressions
            include("expr"),
        ],
        "extra-statements": [],
        "decorator": [
            include("magicfuncs"),
            include("magicvars"),
            include("builtins"),
            (rf"({uni_name})(?=\s|\#)", bygroups(Name.Other.FunctionCall)),
            include("expr"),
            (r"(\#.*)?(?=\n)", bygroups(Comment.Single), "#pop"),
        ],
        "expr": [
            # allow comments
            (r"#.*$", Comment.Single),
            # properly track parentheses and brackets
            (r"\(", Punctuation, "parens"),
            (r"\[", Punctuation, "brackets"),
            (r"\{", Punctuation, "curly-brackets"),
            # the following is copied from PythonLexer
            # raw f-strings
            (
                '(?i)(rf|fr)(""")',
                bygroups(String.Affix, String.Double),
                combined("rfstringescape", "tdqf"),
            ),
            (
                "(?i)(rf|fr)(''')",
                bygroups(String.Affix, String.Single),
                combined("rfstringescape", "tsqf"),
            ),
            (
                '(?i)(rf|fr)(")',
                bygroups(String.Affix, String.Double),
                combined("rfstringescape", "dqf"),
            ),
            (
                "(?i)(rf|fr)(')",
                bygroups(String.Affix, String.Single),
                combined("rfstringescape", "sqf"),
            ),
            # non-raw f-strings
            (
                '([fF])(""")',
                bygroups(String.Affix, String.Double),
                combined("fstringescape", "tdqf"),
            ),
            (
                "([fF])(''')",
                bygroups(String.Affix, String.Single),
                combined("fstringescape", "tsqf"),
            ),
            (
                '([fF])(")',
                bygroups(String.Affix, String.Double),
                combined("fstringescape", "dqf"),
            ),
            (
                "([fF])(')",
                bygroups(String.Affix, String.Single),
                combined("fstringescape", "sqf"),
            ),
            # raw bytes and strings
            ('(?i)(rb|br|r)(""")', bygroups(String.Affix, String.Double), "tdqs"),
            ("(?i)(rb|br|r)(''')", bygroups(String.Affix, String.Single), "tsqs"),
            ('(?i)(rb|br|r)(")', bygroups(String.Affix, String.Double), "dqs"),
            ("(?i)(rb|br|r)(')", bygroups(String.Affix, String.Single), "sqs"),
            # non-raw strings
            (
                '([uU]?)(""")',
                bygroups(String.Affix, String.Double),
                combined("stringescape", "tdqs"),
            ),
            (
                "([uU]?)(''')",
                bygroups(String.Affix, String.Single),
                combined("stringescape", "tsqs"),
            ),
            (
                '([uU]?)(")',
                bygroups(String.Affix, String.Double),
                combined("stringescape", "dqs"),
            ),
            (
                "([uU]?)(')",
                bygroups(String.Affix, String.Single),
                combined("stringescape", "sqs"),
            ),
            # non-raw bytes
            (
                '([bB])(""")',
                bygroups(String.Affix, String.Double),
                combined("bytesescape", "tdqs"),
            ),
            (
                "([bB])(''')",
                bygroups(String.Affix, String.Single),
                combined("bytesescape", "tsqs"),
            ),
            (
                '([bB])(")',
                bygroups(String.Affix, String.Double),
                combined("bytesescape", "dqs"),
            ),
            (
                "([bB])(')",
                bygroups(String.Affix, String.Single),
                combined("bytesescape", "sqs"),
            ),
            # mostly copied from PythonLexer, with changes noted
            (r"[^\S\n]+", Text),
            include("numbers"),
            (r"\b(lambda)\b", Keyword, "lambda"),  # as above, parse lambdas in detail
            (r"!=|==|<<|>>|:=|[-~+/*%=<>&^|@]", Operator),  # include @ but not .
            (r"\.\.\.", Keyword.Constant),  # recognize ellipsis
            (r"\.", Operator.Member),  # special token for member access
            (r"[:,;]", Punctuation),  # exclude brackets (handled above)
            (r"(in|is|and|or|not)\b", Operator.Word),
            include("extra-exprs"),  # hook for subclasses
            include("expr-keywords"),
            (rf"\b(?={uni_name}\s*\()", Text, "function-call"),  # parse function calls
            include("builtins"),
            include("magicfuncs"),
            include("magicvars"),
            (r"(\\\n)", Punctuation),  # extra rule for line continuations
            include("name"),
        ],
        "parens": [
            (r"\)", Punctuation, "#pop"),
            (r"\n", Whitespace),
            include("expr"),
        ],
        "brackets": [
            (r"\]", Punctuation, "#pop"),
            (r"\n", Whitespace),
            include("expr"),
        ],
        "curly-brackets": [
            (r"\}", Punctuation, "#pop"),
            (r"\n", Whitespace),
            include("expr"),
        ],
        "extra-exprs": [],
        "funcname": [
            # parse not only the name, but the list of parameters
            include("magicfuncs"),
            (uni_name, Name.Function),
            (r"\\\n", Punctuation),
            (r"\s+", Whitespace),
            (r"\(", Punctuation, "params"),  # start param list
            (r"->", Punctuation, "return-annotation"),
            (r":", Punctuation, "#pop"),
        ],
        "params": [
            # Use a specific token for parameter names
            (r"(/)(?:(,)|(?=\s*\)))", bygroups(Operator, Punctuation)),
            (r"(\*)(\s*)(,)", bygroups(Operator, Whitespace, Punctuation)),
            (r"\*\*|\*", Operator),
            (
                rf"""(?x)
                (?: (self|cls) | ({uni_name})) (\s*)
                (?: (,) | (?=[)#\n=]))""",
                bygroups(
                    Name.Variable.Parameter.Magic,
                    Name.Variable.Parameter,
                    Whitespace,
                    Punctuation,
                ),
            ),
            (r"#.*$", Comment.Single),
            (r"\\\n", Punctuation),
            (r"=", Operator, "default"),
            (
                rf"\b({uni_name})(\s*)(:)",
                bygroups(Name.Variable.Parameter, Whitespace, Punctuation),
                "annotated-param",
            ),
            (r"\s+", Whitespace),
            (r"\)", Punctuation, "#pop"),
        ],
        "default": [
            (r"\s+", Whitespace),
            (r"(,)|(?=\))", bygroups(Punctuation), "#pop"),
            include("expr"),
        ],
        "annotated-param": [
            (r"\s+", Whitespace),
            (r"(,)|(?=\))", bygroups(Punctuation), "#pop"),
            include("expr"),
            (r"=(?!=)", Operator),
        ],
        "return-annotation": [
            (r"\s+", Whitespace),
            (r"(?=:)", Text, "#pop"),
            include("expr"),
        ],
        "lambda": [
            # Similar to 'params' above
            (r"\*\*|\*", Operator),
            (
                rf"({uni_name})(\s*)(?:(,)|(?=:|$))",
                bygroups(Name.Variable.Parameter, Whitespace, Punctuation),
            ),
            (r"#.*$", Comment.Single),
            (
                rf"\b({uni_name})(\s*)(=)",
                bygroups(Name.Variable.Parameter, Whitespace, Operator),
                "lambda-default",
            ),
            (r"\s+", Whitespace),
            (r"\\\n", Punctuation),
            (r"(:)|(\n)", bygroups(Punctuation, Whitespace), "#pop"),
        ],
        "lambda-default": [
            (r"(,)|(?=:|$)", bygroups(Punctuation), "#pop"),
            include("expr"),
        ],
        "function-call": [
            # Use a specific token for called functions, and parse arguments
            (r"\)", Punctuation, "#pop"),
            (r"\s+", Whitespace),
            (r"(self|cls)\b", Name.Builtin.Pseudo),
            include("magicfuncs"),
            include("magicvars"),
            include("builtins"),
            (rf"\b{uni_name}\b", Name.Other.FunctionCall),
            (r"\(", Punctuation, "function-arguments"),
        ],
        "function-arguments": [
            # Label names of default parameters as parameters
            (r"(?=\))(?!\)\s*\()", Text, "#pop"),
            (r",", Punctuation),
            (r"\*\*|\*", Operator),
            (
                rf"\b({uni_name})(\s*)(=)(?!=)",
                bygroups(Name.Variable.Parameter, Whitespace, Operator),
            ),
            include("expr"),
            (
                r"(\s*)(\))(\s*)(\()",
                bygroups(Whitespace, Punctuation, Whitespace, Punctuation),
            ),
            (r"\s+", Whitespace),
        ],
    }


class ScenicLexer(BetterPythonLexer):
    """Lexer for Scenic code."""

    name = "Scenic"
    aliases = ["scenic"]
    filenames = ["*.scenic"]
    alias_filenames = ["*.sc"]
    mimetypes = ["application/x-scenic", "text/x-scenic"]
    url = "https://scenic-lang.readthedocs.org/"

    uni_name = PythonLexer.uni_name
    obj_name = rf"(?:(ego)|({uni_name}))"

    simple_spec = words(
        (
            "at",
            "offset by",
            "offset along",
            "left of",
            "right of",
            "ahead of",
            "above",
            "below",
            "behind",
            "beyond",
            "visible from",
            "visible",
            "not visible from",
            "not visible",
            "in",
            "on",
            "contained in",
            "following",
            "facing toward",
            "facing away from",
            "facing directly toward",
            "facing directly away from",
            "facing",
            "apparently facing",
        )
    ).get()
    specifier = rf"""
        \b(?<!\.) (?:
            (with) (\s+) ({uni_name})
            | ({simple_spec})
        )\b"""

    simple_statements = words(
        (
            "simulator",
            "param",
            "require",
            "require monitor",
            "terminate when",
            "terminate after",
            "mutate",
            "record initial",
            "record final",
            "record",
            "take",
            "wait",
            "terminate",
            "do choose",
            "do shuffle",
            "do",
            "abort",
            "interrupt when",
        ),
        suffix=r"\b(?!\s*[:.;,])",
    ).get()

    constructible_types = ("Point", "OrientedPoint", "Object")
    primitive_types = (
        "Vector",
        "Orientation",
        "VectorField",
        "PolygonalVectorField",
        "Shape",
        "MeshShape",
        "BoxShape",
        "CylinderShape",
        "ConeShape",
        "SpheroidShape",
        "Region",
        "PointSetRegion",
        "RectangularRegion",
        "CircularRegion",
        "SectorRegion",
        "PolygonalRegion",
        "PolylineRegion",
        "PathRegion",
        "MeshVolumeRegion",
        "MeshSurfaceRegion",
        "BoxRegion",
        "SpheroidRegion",
        "Workspace",
        "Range",
        "DiscreteRange",
        "Options",
        "Discrete",
        "Uniform",
        "Normal",
        "TruncatedNormal",
        "VerifaiParameter",
        "VerifaiRange",
        "VerifaiDiscreteRange",
        "VerifaiOptions",
    )
    exceptions = ("GuardViolation", "PreconditionViolation", "InvariantViolation")
    builtin_names = (
        "everywhere",
        "nowhere",
        "resample",
        "localPath",
        "verbosePrint",
        "simulation",
        "sin",
        "cos",
        "hypot",
    )
    magic_names = ("ego", "workspace", "globalParameters")
    builtin_classes = words(constructible_types, suffix=r"\b").get()

    tokens = {
        "extra-statements": [
            # Scenario and other block definitions
            (
                rf"^(\s*)(scenario|behavior|monitor)(\s+)",
                bygroups(Whitespace, Keyword, Text),
                "scenario-behavior",
            ),
            # Model statement (which highlights the module name specially, like import)
            (r"^(\s*)(model)(\s+)", bygroups(Whitespace, Keyword, Whitespace), "model"),
            # Override statement (which ends with a list of specifiers)
            (
                rf"""(?x)
                ^(\s*) (override)
                (\s+) {obj_name} ([^\S\n]*)
                (?={specifier})""",
                bygroups(
                    Whitespace, Keyword, Whitespace, Name.Variable.Magic, Name, Whitespace
                ),
                "specifier-start",
            ),
            # Param statement
            (
                r"^(\s*)(param)(\s+)",
                bygroups(Whitespace, Keyword, Whitespace),
                "param-statement",
            ),
            # Require statement (within which we can have temporal operators)
            (
                r"^(\s*)(require)(\s+)(?!monitor)",
                bygroups(Whitespace, Keyword, Whitespace),
                "require-statement",
            ),
            # Keywords that can only occur at the start of a line
            (rf"^(\s*)({simple_statements})", bygroups(Whitespace, Keyword)),
            (
                r"^(\s*)(try|setup|compose|precondition|invariant)(:)",
                bygroups(Whitespace, Keyword, Punctuation),
            ),
            # Keywords that can only occur at the end of a line
            (r"(?<!\.)(seconds|steps)\b(?=\s*(\#.*)?$)", Keyword),
            # Property declarations
            (
                rf"""(?x)
                ^(\s+)
                (?!else|except|finally) ({uni_name})
                (\s*)
                (?: (\[) (?= [\w\s,]* \] \s* : \s* [^\s\#])
                    | (?= : \s* [^\s\#]))""",
                bygroups(Whitespace, Name.Variable.Instance, Whitespace, Punctuation),
                "property-declaration",
            ),
        ],
        "scenario-behavior": [
            (uni_name, Name.Class),
            (r"\\\n", Punctuation),
            (r"\s+", Whitespace),
            (r"\(", Punctuation, "params"),
            (r":", Punctuation, "#pop"),
        ],
        "model": [
            (r"\.", Name.Namespace),
            (uni_name, Name.Namespace),
            default("#pop"),
        ],
        "param-statement": [
            (
                rf"""(?x)
                ({uni_name} | '[^\\\'"%{{\n]+' | "[^\\\'"%{{\n]+")
                (\s*) (=) (\s*)""",
                bygroups(Name.Variable.Global, Whitespace, Operator, Whitespace),
            ),
            include("expr"),
            (r"(\#.*)?\n", Comment.Single, "#pop"),
        ],
        "require-statement": [
            (
                words(("always", "eventually", "next", "until", "implies"), suffix=r"\b"),
                Keyword,
            ),
            (r"\(", Punctuation, "require-parens"),
            include("expr"),
            (r"(\#.*)?\n", Comment.Single, "#pop"),
        ],
        "require-parens": [
            (r"\)", Punctuation, "#pop"),
            (r"\n", Whitespace),
            include("require-statement"),
        ],
        "property-declaration": [
            (words(("additive", "dynamic", "final"), suffix=r"\b"), Keyword.Pseudo),
            (rf"{uni_name}\b", Text),
            (r"(,)(\s*)", bygroups(Punctuation, Whitespace)),
            (r"(\])?(\s*)(:)", bygroups(Punctuation, Whitespace, Punctuation), "#pop"),
        ],
        "specifier-start": [
            (
                r"(?x)" + specifier,
                bygroups(
                    Keyword.Specifier,
                    Whitespace,
                    Name.Variable.Instance,
                    Keyword.Specifier,
                ),
                "specifier-inner",
            ),
            default("#pop"),
        ],
        "specifier-inner": [
            (
                r"""(?x)
             (,) (\s*) (?: (?: (\#.*)? | (\\)) (\n\s*))?
             | (\#.*)? (?=[)}\]\n])""",
                bygroups(
                    Punctuation,
                    Whitespace,
                    Comment.Single,
                    Punctuation,
                    Whitespace,
                    Comment.Single,
                ),
                "#pop",
            ),
            include("expr"),
        ],
        "extra-exprs": [
            # Prefix and infix operators
            (
                words(
                    (
                        "relative heading of",
                        "apparent heading of",
                        "distance from",
                        "distance to",
                        "distance past",
                        "angle from",
                        "angle to",
                        "altitude from",
                        "altitude to",
                        "can see",
                        "at",
                        "relative to",
                        "offset by",
                        "offset along",
                        "visible",
                        "not visible",
                        "front of",
                        "back of",
                        "left of",
                        "right of",
                        "front left of",
                        "front right of",
                        "back left of",
                        "back right of",
                    ),
                    prefix=r"(?<!\.)",
                    suffix=r"\b(?!\s*(?:[)}\]=:.;,#]|$))",
                ),
                Operator.Word,
            ),
            # Postfix operators
            (words(("deg",), prefix=r"(?<!\.)", suffix=r"\b"), Operator.Word),
            # Keywords that can occur anywhere (w.r.t. our simple analysis)
            (
                words(
                    ("initial scenario", "until", "as", "to", "by", "from"),
                    prefix=r"(?<!\.)",
                    suffix=r"\b",
                ),
                Keyword,
            ),
            # Instance creations
            (
                rf"""(?x)
                (new) ([^\S\n]+)
                (?:{builtin_classes}|({uni_name})) ([^\S\n]*)""",
                bygroups(
                    Keyword,
                    Whitespace,
                    Name.Builtin.Instance,
                    Name.Class.Instance,
                    Whitespace,
                ),
                "specifier-start",
            ),
        ],
        "builtins": [
            inherit,
            (
                words(
                    constructible_types + primitive_types + builtin_names,
                    prefix=r"(?<!\.)",
                    suffix=r"\b",
                ),
                Name.Builtin,
            ),
            (words(magic_names, prefix=r"(?<!\.)", suffix=r"\b"), Name.Variable.Magic),
            (words(exceptions, prefix=r"(?<!\.)", suffix=r"\b"), Name.Exception),
        ],
    }

    def analyse_text(text):
        score = 0
        if "new Object" in text:
            score += 0.2
        if "ego = new " in text:
            score += 0.2
        if " facing " in text:
            score += 0.2
        if "param " in text:
            score += 0.1
        if "require " in text:
            score += 0.1
        return score


class ScenicSnippetLexer(ScenicLexer):
    """Variant ScenicLexer for code snippets rather than complete programs.

    Specifically, this lexer formats syntactic variables of the form "{name}"
    as "name" italicized.
    """

    name = "Scenic (snippets)"
    aliases = ["scenic-snippet"]
    filenames = []
    alias_filenames = []
    mimetypes = []

    tokens = {
        "expr": [
            (r"\{([\w.]*)\}", bygroups(Generic.Emph)),
            inherit,
        ],
    }

    def analyse_text(text):
        return 0.0


class PythonSnippetLexer(BetterPythonLexer):
    """Variant PythonLexer for code snippets rather than complete programs.

    Specifically, this lexer formats syntactic variables of the form "{name}"
    as "name" italicized.
    """

    name = "Python (snippets)"
    aliases = ["python-snippet"]
    filenames = []
    alias_filenames = []
    mimetypes = []

    tokens = {
        "expr": [
            (r"\{([\w.]*)\}", bygroups(Generic.Emph)),
            inherit,
        ],
    }

    def analyse_text(text):
        return 0.0


class ScenicSpecifierLexer(ScenicSnippetLexer):
    """Further variant lexer for specifiers at the top level."""

    name = "Scenic (specifier snippets)"
    aliases = ["scenic-specifier"]

    tokens = {
        "root": [
            (
                r"(?x)" + ScenicLexer.specifier,
                bygroups(
                    Keyword.Specifier,
                    Whitespace,
                    Name.Variable.Instance,
                    Keyword.Specifier,
                ),
                "specifier-inner",
            ),
            (r"with", Keyword.Specifier),  # special case for incomplete 'with'
        ],
    }


class ScenicRequirementLexer(ScenicSnippetLexer):
    """Further variant lexer for requirements at the top level."""

    name = "Scenic (requirement snippets)"
    aliases = ["scenic-requirement"]

    tokens = {
        "root": [
            include("require-statement"),
        ],
    }


class ScenicPropertyLexer(RegexLexer):
    """Silly lexer to color property names consistently with the real lexer."""

    name = "Scenic (properties)"
    aliases = ["scenic-property"]

    tokens = {
        "root": [
            (ScenicLexer.uni_name, Name.Variable.Instance),
        ],
    }

    def analyse_text(text):
        return 0.0


class ScenicGrammarLexer(RegexLexer):
    """Lexer for the grammar notation used in the Scenic docs."""

    name = "Scenic Grammar"
    aliases = ["scenic-grammar"]

    tokens = {
        "root": [
            (
                r"(<)([ -;=?-~]+)(>)",
                bygroups(Punctuation, Name.Class.Grammar, Punctuation),
            ),
            (r"[\[\]|/*+]", Operator),
            (r"\#.*\n", Comment.Single),
            (r"[^<>\[\]|*+#]+", Text),  # for performance
            (r".", Text),
        ]
    }

    def analyse_text(text):
        return 0.0


class ScenicStyle(Style):
    """A style providing specialized highlighting for the Scenic language.

    The color scheme is a loose hybrid of that used in the Scenic papers and
    the 'Mariana' color scheme from Sublime Text. The chosen colors all have a
    contrast ratio of at least 4.5:1 against the background color, per the
    W3C's Web Content Accessibility Guidelines.
    """

    background_color = "#f8f8f8"

    # Inherit styles from the default style for any token types we don't use
    styles = DefaultStyle.styles

    # fmt: off
    styles.update({
        Punctuation:                    "#656E81",
        Comment:                        "italic #707274",

        Keyword:                        "#6852BB",
        Keyword.Specifier:              "#146666",
        Keyword.Constant:               "italic #4042B3",

        Operator:                       "#C7161C",
        Operator.Word:                  "#992024",
        Operator.Member:                "#656E81",

        Name:                           "#404040",
        Name.Builtin:                   "italic #1773cf",
        Name.Builtin.Pseudo:            "#A66214",
        Name.Builtin.Instance:          "bold #3C8031",
        Name.Exception:                 "italic #CB3F38",
        Name.Variable.Parameter:        "#A66214",
        Name.Namespace:                 "#0077A6",
        Name.Function:                  "#0077A6",
        Name.Function.Magic:            "italic #1773cf",
        Name.Other.FunctionCall:        "#1773cf",
        Name.Class:                     "bold #0077A6",
        Name.Class.Grammar:             "nobold italic",
        Name.Class.Instance:            "#3C8031",
        Name.Variable:                  "#B34024",
        Name.Variable.Parameter.Magic:  "italic",
        Name.Variable.Magic:            "italic #B32483",

        String:                         "#34822B",
        String.Affix:                   "#6852BB",
        String.Interpol:                "bold #B32483",
        String.Escape:                  "#B32483",
        Number:                         "#4042B3",

        Error:                          "#FFFFFF bg:#FF0000",
    })
    # fmt: on


class PegenLexer(BetterPythonLexer):
    """Lexer for Pegen grammars."""

    name = "Pegen"
    aliases = ["pegen"]
    filenames = ["*.gram"]
    mimetypes = ["application/x-pegen", "text/x-pegen"]
    url = "https://we-like-parsers.github.io/pegen/"

    uni_name = PythonLexer.uni_name
    quotes = "'''|\"\"\""
    constants = (
        # Token types
        "ENDMARKER",
        "NAME",
        "NUMBER",
        "STRING",
        "NEWLINE",
        "INDENT",
        "DEDENT",
        "LPAR",
        "RPAR",
        "LSQB",
        "RSQB",
        "COLON",
        "COMMA",
        "SEMI",
        "PLUS",
        "MINUS",
        "STAR",
        "SLASH",
        "VBAR",
        "AMPER",
        "LESS",
        "GREATER",
        "EQUAL",
        "DOT",
        "PERCENT",
        "LBRACE",
        "RBRACE",
        "EQEQUAL",
        "NOTEQUAL",
        "LESSEQUAL",
        "GREATEREQUAL",
        "TILDE",
        "CIRCUMFLEX",
        "LEFTSHIFT",
        "RIGHTSHIFT",
        "DOUBLESTAR",
        "PLUSEQUAL",
        "MINEQUAL",
        "STAREQUAL",
        "SLASHEQUAL",
        "PERCENTEQUAL",
        "AMPEREQUAL",
        "VBAREQUAL",
        "CIRCUMFLEXEQUAL",
        "LEFTSHIFTEQUAL",
        "RIGHTSHIFTEQUAL",
        "DOUBLESTAREQUAL",
        "DOUBLESLASH",
        "DOUBLESLASHEQUAL",
        "AT",
        "ATEQUAL",
        "RARROW",
        "ELLIPSIS",
        "COLONEQUAL",
        "OP",
        "AWAIT",
        "ASYNC",
        "TYPE_IGNORE",
        "TYPE_COMMENT",
        "ERRORTOKEN",
        "COMMENT",
        "NL",
        "ENCODING",
        # Misc others
        "SOFT_KEYWORD",
        "$",
    )

    tokens = {
        "root": [
            (r"\n", Whitespace),
            (r"#.*$", Comment.Single),
            # Metas with special formatting
            (
                rf"(@class)([^\S\n]+)({uni_name})",
                bygroups(Keyword, Whitespace, Name.Class),
            ),
            (
                rf"(@(?:header|subheader|trailer))([^\S\n]*)({quotes})((?:.|\n)*)(\3)",
                bygroups(
                    Keyword,
                    Whitespace,
                    String.Interpol,
                    using(BetterPythonLexer),
                    String.Interpol,
                ),
            ),
            # Generic rules in case additional metas are added
            (
                rf"(@{uni_name})(?:([^\S\n]+)({uni_name}))?(\n)",
                bygroups(Keyword, Whitespace, Name, Whitespace),
            ),
            (
                rf"(@{uni_name})([^\S\n]*)('''|')((?:.|\n)*?\3)",
                bygroups(Keyword, Whitespace, String.Single, String.Single),
            ),
            (
                rf'(@{uni_name})([^\S\n]*)("""|")((?:.|\n)*?\3)',
                bygroups(Keyword, Whitespace, String.Double, String.Double),
            ),
            # Start of a rule
            (uni_name, Name.Class, "rulestart"),
        ],
        "rulestart": [
            (r"\[", Punctuation, "ruletype"),
            (r"(\()(memo)(\))", bygroups(Punctuation, Keyword, Punctuation)),
            (r":", Punctuation, "rulebody"),
            (r"[^\S\n]+", Whitespace),
        ],
        "ruletype": [
            include("expr"),
            (r"\]", Punctuation, "#pop"),
        ],
        "rulebody": [
            (r"(\s*)(#.*)?(\n)", bygroups(Whitespace, Comment, Whitespace), "altlist"),
            include("gram-expr"),
        ],
        "altlist": [
            (r"(\s+)(\|)", bygroups(Whitespace, Operator), "gram-expr"),
            (r"(\s+)(#.*\n)", bygroups(Whitespace, Comment.Single)),
            default("#pop:3"),
        ],
        "gram-expr": [
            (r"\(", Punctuation, "gram-parens"),
            (r"\[", Operator, "gram-brackets"),
            (r"\{", String.Interpol, "gram-curly-brackets"),
            (words(constants, suffix=r"\b"), Keyword.Constant),
            (r"[^\S\n]+", Whitespace),
            (r"#.*$", Comment.Single),
            (r"[|?*+.&!~]", Operator),
            (r"'[^']*'", String.Single),
            (r'"[^"]*"', String.Double),
            (rf"({uni_name})(=)", bygroups(Name.Variable.Class, Operator)),
            (uni_name, Name),
            (r"\n", Whitespace, "#pop"),
        ],
        "gram-parens": [
            (r"\)", Punctuation, "#pop"),
            (r"\n", Whitespace),
            include("gram-expr"),
        ],
        "gram-brackets": [
            (r"\]", Operator, "#pop"),
            (r"\n", Whitespace),
            include("gram-expr"),
        ],
        "gram-curly-brackets": [
            include("expr"),
            (r"\}", String.Interpol, "#pop"),
            (r"\n", Whitespace),
        ],
        "expr-keywords": [
            inherit,
            (words(("LOCATIONS", "UNREACHABLE"), suffix=r"\b"), Keyword.Constant),
        ],
    }

    def analyse_text(text):
        if "@subheader" in text:
            return 1.0
        elif "@class" in text:
            return 0.5
        else:
            return 0.0



"""Car models built into Webots."""

from dataclasses import dataclass


@dataclass(frozen=True)
class CarModel:
    name: str
    width: float
    length: float
    height: float


carModels = {
    CarModel("BmwX5", 2, 4.75, 0),
    CarModel("CitroenCZero", 1.5, 3.5, 0),
    CarModel("LincolnMKZ", 1.9, 5, 0),
    CarModel("RangeRoverSportSVR", 1.9, 4.5, 0),
    CarModel("ToyotaPrius", 1.9, 4.5, 0),
    CarModel("Bus", 3, 10, 0),
    CarModel("Truck", 2.7, 14.5, 0),
    CarModel("Tractor", 2, 3, 0),
    CarModel("MotorBikeSimple", 1, 2.5, 0),
}
modelWithName = {model.name: model for model in carModels}

smallVehicles = {
    modelWithName[name]
    for name in {
        "BmwX5",
        "CitroenCZero",
        "LincolnMKZ",
        "RangeRoverSportSVR",
        "ToyotaPrius",
    }
}



"""Support for checking Scenic types.

This module provides a system for checking that values passed to Scenic operators and
functions have the expected types. The top-level function `toTypes` and its
specializations `toType`, `toVector`, `toScalar`, etc. can also *coerce* closely-related
types into the desired type in some cases. For lazily-evaluated values (random values and
delayed arguments of specifiers), it may not be possible to determine the type at object
creation time: in such cases these functions return a lazily-evaluated object that
performs the type check either during specifier resolution or sampling as needed.

In general, the only objects which are coercible to a type T are instances of that type,
together with `Distribution` objects whose **_valueType** is a type coercible to T (and
therefore whose sampled value can be coerced to T). However, we also have the following
exceptional rules:

    * Coercible to a scalar (type `float`):
        * Instances of `numbers.Real` (coerced by calling `float` on them);
          this includes NumPy types such as `numpy.single`
    * Coercible to a heading (type `Heading`):
        * Anything coercible to a scalar
        * Any type with a **toHeading** method (including `OrientedPoint`)
    * Coercible to a vector (type `Vector`):
        * Tuples and lists of length 2 or 3
        * Any type with a **toVector** method (including `Point`)
    * Coercible to a `Behavior`:
        * Subclasses of `Behavior` (coerced by calling them with no arguments)
        * `None` (considered to have type `Behavior` for convenience)
"""

import inspect
import numbers
import sys
import typing
from typing import get_args as get_type_args, get_origin as get_type_origin

from scenic.core.distributions import (
    Distribution,
    RejectionException,
    StarredDistribution,
    TupleDistribution,
    distributionFunction,
    supportInterval,
    toDistribution,
)
from scenic.core.errors import saveErrorLocation
from scenic.core.lazy_eval import (
    DelayedArgument,
    needsLazyEvaluation,
    requiredProperties,
    valueInContext,
)

## Basic types


class Heading(float):
    """Dummy class used as a target for type coercions to headings."""

    pass


def underlyingType(thing):
    """What type this value ultimately evaluates to, if we can tell."""
    if isinstance(thing, Distribution):
        return thing._valueType
    elif isinstance(thing, TypeChecker) and len(thing.types) == 1:
        return thing.types[0]
    elif isinstance(thing, TypeChecker):
        return object
    else:
        return type(thing)


def isA(thing, ty):
    """Is this guaranteed to evaluate to a member of the given Scenic type?"""
    # TODO: Remove this hack once type system is smarter.
    if not isinstance(underlyingType(thing), type):
        return False

    return issubclass(underlyingType(thing), ty)


def unifyingType(opts):  # TODO improve?
    """Most specific type unifying the given values."""
    # Gather underlying types of all options
    types = []
    for opt in opts:
        if isinstance(opt, StarredDistribution):
            ty = underlyingType(opt)
            typeargs = get_type_args(ty)
            if typeargs == ():
                types.append(ty)
            else:
                for ty in typeargs:
                    if ty is not Ellipsis:
                        types.append(ty)
        else:
            types.append(underlyingType(opt))
    # Compute unifying type
    return unifierOfTypes(types)


def unifierOfTypes(types):
    """Most specific type unifying the given types."""
    # If all types are equal, unifier is trivial
    first = types[0]
    if all(ty == first for ty in types):
        return first
    # Otherwise, erase type parameters which we don't understand
    simpleTypes = []
    for ty in types:
        origin = get_type_origin(ty)
        simpleTypes.append(origin if origin else ty)
    # Scalar types unify to float
    if all(issubclass(ty, numbers.Real) for ty in simpleTypes):
        return float
    # For all other types, find first common ancestor in MRO
    # (ignoring type parameters and skipping ABCs)
    mro = inspect.getmro(simpleTypes[0])
    for ancestor in mro:
        if inspect.isabstract(ancestor):
            continue
        if all(issubclass(ty, ancestor) for ty in simpleTypes):
            return ancestor
    raise AssertionError(f"broken MRO for types {types}")


## Type coercions (for internal use -- see the type checking API below)


def canCoerceType(typeA, typeB):
    """Can values of typeA be coerced into typeB?"""
    originA = get_type_origin(typeA)
    if originA is typing.Union:
        # only raise an error now if none of the possible types will work;
        # we'll do more careful checking at runtime
        return any(canCoerceType(ty, typeB) for ty in get_type_args(typeA))
    elif originA:
        # erase type parameters which we don't know how to use
        typeA = originA
    if typeB == float:
        return issubclass(typeA, numbers.Real)
    elif typeB == Heading:
        from scenic.core.vectors import Orientation

        return (
            canCoerceType(typeA, float)
            or hasattr(typeA, "toHeading")
            or issubclass(typeA, Orientation)
        )
    elif hasattr(typeB, "_canCoerceType"):
        return typeB._canCoerceType(typeA)
    else:
        return issubclass(typeA, typeB)


def canCoerce(thing, ty, exact=False):
    """Can this value be coerced into the given type?"""
    tt = underlyingType(thing)
    if canCoerceType(tt, ty):
        return True
    elif (not exact) and isinstance(thing, Distribution):
        return True  # fall back on type-checking at runtime
    else:
        return False


def coerce(thing, ty, error="wrong type"):
    """Coerce something into the given type.

    Used internally by `toType`, etc.; this function should not otherwise be
    called directly.
    """
    assert canCoerce(thing, ty), (thing, ty)

    # If we are in any of the exceptional cases (see the module documentation above),
    # select the appropriate helper function for performing the coercion.
    realType = ty
    if ty == float:
        coercer = coerceToFloat
    elif ty == Heading:
        coercer = coerceToHeading
        ty = numbers.Real
        realType = float
    else:
        coercer = getattr(ty, "_coerce", None)

    # Special case: can coerce TupleDistribution directly to Vector.
    # (all other distributions require the usual checking below)
    from scenic.core.vectors import Vector

    if isinstance(thing, TupleDistribution) and ty is Vector:
        length = len(thing)
        if not (2 <= length <= 3):
            msg = f"expected vector, got {thing.builder.__name__} of length {length}"
            raise TypeError(f"{error} ({msg})")
        return Vector(*thing)

    if isinstance(thing, Distribution):
        # This is a random value. If we can prove that it will always have the desired
        # type, we return it unchanged; otherwise we wrap it in a TypecheckedDistribution
        # for sample-time typechecking.
        vt = thing._valueType
        origin = get_type_origin(vt)
        if origin is typing.Union:
            possibleTypes = get_type_args(vt)
        elif origin:
            # Erase type parameters which we don't understand
            possibleTypes = (origin,)
        else:
            possibleTypes = (vt,)
        if all(issubclass(possible, ty) for possible in possibleTypes):
            return thing  # no coercion necessary
        else:
            return TypecheckedDistribution(thing, realType, error, coercer=coercer)
    elif coercer:
        # The destination type has special coercion rules: call the appropriate helper.
        try:
            return coercer(thing)
        except CoercionFailure as e:
            raise TypeError(f"{error} ({e.args[0]})") from None
    else:
        # Only instances of the destination type can be coerced into it; since coercion
        # is possible, we must have such an instance, and can return it unchanged.
        return thing


class CoercionFailure(Exception):
    """Raised by coercion functions when coercion is impossible.

    Only used internally; will be converted to a parse error for reporting to
    the user.
    """

    pass


def coerceToFloat(thing) -> float:
    return float(thing)


def coerceToHeading(thing) -> Heading:
    from scenic.core.vectors import Orientation

    if isinstance(thing, Orientation):
        return thing.yaw

    h = thing.toHeading() if hasattr(thing, "toHeading") else float(thing)
    return h


class TypecheckedDistribution(Distribution):
    """Distribution which typechecks its value at sampling time.

    Only for internal use by the typechecking system; introduced by `coerce` when it is
    unable to guarantee that a random value will have the correct type after sampling.
    Note that the type check is not a purely passive operation, and may actually
    transform the sampled value according to the coercion rules above (e.g. a sampled
    `Point` will be converted to a `Vector` in a context which expects the latter).
    """

    _deterministic = True

    def __init__(self, dist, ty, errorMessage, coercer=None):
        super().__init__(dist, valueType=ty)
        self._dist = dist
        self._errorMessage = errorMessage
        self._coercer = coercer
        if not coercer:
            # Erase any type parameters, which we don't attempt to check
            origin = get_type_origin(ty)
            self._checkType = origin if origin else ty
        self._loc = saveErrorLocation()

    def sampleGiven(self, value):
        val = value[self._dist]
        suffix = None
        # Attempt to coerce sampled value into the desired type.
        if self._coercer:
            if canCoerceType(type(val), self._valueType):
                try:
                    return self._coercer(val)
                except CoercionFailure as e:
                    suffix = f" ({e.args[0]})"
        elif isinstance(val, self._checkType):
            return val
        # Coercion failed, so we have a type error.
        if suffix is None:
            suffix = f" (expected {self._valueType.__name__}, got {type(val).__name__})"
        exc = TypeError(self._errorMessage + suffix)
        exc._scenic_location = self._loc
        raise exc

    def conditionTo(self, value):
        self._dist.conditionTo(value)

    def supportInterval(self):
        return supportInterval(self._dist)

    def __repr__(self):
        return f"TypecheckedDistribution({self._dist!r}, {self._valueType!r})"


def coerceToAny(thing, types, error):
    """Coerce something into any of the given types, raising an error if impossible.

    Only for internal use by the typechecking system; called from `toTypes`.

    Raises:
        TypeError: if it is impossible to coerce the value into any of the types.
    """
    for ty in types:
        if canCoerce(thing, ty):
            return coerce(thing, ty, error)
    from scenic.syntax.veneer import verbosePrint

    verbosePrint(
        f"Failed to coerce {thing} of type {underlyingType(thing)} to {types}",
        file=sys.stderr,
    )
    raise TypeError(error)


## Top-level type checking/conversion API


def toTypes(thing, types, typeError="wrong type"):
    """Convert something to any of the given types, raising an error if impossible.

    Types are tried in the order they are given: the first one compatible with the given
    value is used. Coercions of closely-related types may take place as described in the
    module documentation above.

    If the given value requires lazy evaluation, this function returns a `TypeChecker`
    object that performs the type conversion after specifier resolution.

    Args:
        thing: Value to convert.
        types: Sequence of one or more destination types.
        typeError (str): Message included in exception raised on failure.

    Raises:
        TypeError: if the given value is not one of the given types and cannot
            be converted to any of them.
    """
    thing = toDistribution(thing)
    if needsLazyEvaluation(thing):
        # cannot check the type now; create proxy object to check type after evaluation
        return TypeChecker(thing, types, typeError)
    else:
        return coerceToAny(thing, types, typeError)


def toType(thing, ty, typeError="wrong type"):
    """Convert something to a given type, raising an error if impossible.

    Equivalent to `toTypes` with a single destination type.
    """
    return toTypes(thing, (ty,), typeError)


def toScalar(thing, typeError="non-scalar in scalar context"):
    """Convert something to a scalar, raising an error if impossible.

    See `toTypes` for details.
    """
    return toType(thing, float, typeError)


def toHeading(thing, typeError="non-heading in heading context"):
    """Convert something to a heading, raising an error if impossible.

    See `toTypes` for details.
    """
    return toType(thing, Heading, typeError)


def toOrientation(thing, typeError="non-orientation in orientation context"):
    """Convert something to an orientation, raising an error if impossible.

    See `toTypes` for details.
    """
    from scenic.core.vectors import Orientation

    return toType(thing, Orientation, typeError)


def toVector(thing, typeError="non-vector in vector context"):
    """Convert something to a vector, raising an error if impossible.

    See `toTypes` for details.
    """
    from scenic.core.vectors import Vector

    return toType(thing, Vector, typeError)


def evaluateRequiringEqualTypes(func, thingA, thingB, typeError="type mismatch"):
    """Evaluate the func, assuming thingA and thingB have the same type.

    If func produces a lazy value, it should not have any required properties beyond
    those of thingA and thingB.

    Raises:
        TypeError: if thingA and thingB do not have the same type.
    """
    if not needsLazyEvaluation(thingA) and not needsLazyEvaluation(thingB):
        if underlyingType(thingA) is not underlyingType(thingB):
            raise TypeError(typeError)
        return func()
    else:
        # cannot check the types now; create proxy object to check types after evaluation
        return TypeEqualityChecker(func, thingA, thingB, typeError)


## Proxy objects for lazy type checking


class TypeChecker(DelayedArgument):
    """Checks that a given lazy value has one of a given list of types."""

    def __init__(self, arg, types, error):
        def check(context):
            val = arg.evaluateIn(context)
            return coerceToAny(val, types, error)

        super().__init__(requiredProperties(arg), check)
        self.inner = arg
        self.types = types

    def __repr__(self):
        return f"TypeChecker({self.inner!r},{self.types!r})"


class TypeEqualityChecker(DelayedArgument):
    """Evaluates a function after checking that two lazy values have the same type."""

    def __init__(self, func, checkA, checkB, error):
        props = requiredProperties(checkA) | requiredProperties(checkB)

        def check(context):
            ca = valueInContext(checkA, context)
            cb = valueInContext(checkB, context)
            if underlyingType(ca) is not underlyingType(cb):
                raise TypeError(error)
            return valueInContext(func(), context)

        super().__init__(props, check)
        self.inner = func
        self.checkA = checkA
        self.checkB = checkB

    def __repr__(self):
        return f"TypeEqualityChecker({self.inner!r},{self.checkA!r},{self.checkB!r})"


## Utilities


def is_typing_generic(tp):
    """Whether this is a pre-3.9 generic type from the typing module."""
    return isinstance(tp, typing._GenericAlias)



"""Simulator interface for CARLA."""

try:
    import carla
except ImportError as e:
    raise ModuleNotFoundError('CARLA scenarios require the "carla" Python package') from e

import math
import os
import warnings

import scenic.core.errors as errors

if errors.verbosityLevel == 0:  # suppress pygame advertisement at zero verbosity
    os.environ["PYGAME_HIDE_SUPPORT_PROMPT"] = "hide"

import pygame

from scenic.core.simulators import SimulationCreationError
from scenic.domains.driving.simulators import DrivingSimulation, DrivingSimulator
from scenic.simulators.carla.blueprints import oldBlueprintNames
import scenic.simulators.carla.utils.utils as utils
import scenic.simulators.carla.utils.visuals as visuals
from scenic.syntax.veneer import verbosePrint


class CarlaSimulator(DrivingSimulator):
    """Implementation of `Simulator` for CARLA."""

    def __init__(
        self,
        carla_map,
        map_path,
        address="127.0.0.1",
        port=2000,
        timeout=10,
        render=True,
        record="",
        timestep=0.1,
        traffic_manager_port=None,
    ):
        super().__init__()
        verbosePrint(f"Connecting to CARLA on port {port}")
        self.client = carla.Client(address, port)
        self.client.set_timeout(timeout)  # limits networking operations (seconds)
        if carla_map is not None:
            try:
                self.world = self.client.load_world(carla_map)
            except Exception as e:
                raise RuntimeError(f"CARLA could not load world '{carla_map}'") from e
        else:
            if str(map_path).endswith(".xodr"):
                with open(map_path) as odr_file:
                    self.world = self.client.generate_opendrive_world(odr_file.read())
            else:
                raise RuntimeError("CARLA only supports OpenDrive maps")
        self.timestep = timestep

        if traffic_manager_port is None:
            traffic_manager_port = port + 6000
        self.tm = self.client.get_trafficmanager(traffic_manager_port)
        self.tm.set_synchronous_mode(True)

        # Set to synchronous with fixed timestep
        settings = self.world.get_settings()
        settings.synchronous_mode = True
        settings.fixed_delta_seconds = timestep  # NOTE: Should not exceed 0.1
        self.world.apply_settings(settings)
        verbosePrint("Map loaded in simulator.")

        self.render = render  # visualization mode ON/OFF
        self.record = record  # whether to use the carla recorder
        self.scenario_number = 0  # Number of the scenario executed

    def createSimulation(self, scene, *, timestep, **kwargs):
        if timestep is not None and timestep != self.timestep:
            raise RuntimeError(
                "cannot customize timestep for individual CARLA simulations; "
                "set timestep when creating the CarlaSimulator instead"
            )

        self.scenario_number += 1
        return CarlaSimulation(
            scene,
            self.client,
            self.tm,
            self.render,
            self.record,
            self.scenario_number,
            timestep=self.timestep,
            **kwargs,
        )

    def destroy(self):
        super().destroy()
        settings = self.world.get_settings()
        settings.synchronous_mode = False
        settings.fixed_delta_seconds = None
        self.world.apply_settings(settings)
        self.tm.set_synchronous_mode(False)


class CarlaSimulation(DrivingSimulation):
    def __init__(self, scene, client, tm, render, record, scenario_number, **kwargs):
        self.client = client
        self.world = self.client.get_world()
        self.map = self.world.get_map()
        self.blueprintLib = self.world.get_blueprint_library()
        self.tm = tm
        self.render = render
        self.record = record
        self.scenario_number = scenario_number
        self.cameraManager = None

        super().__init__(scene, **kwargs)

    def setup(self):
        weather = self.scene.params.get("weather")
        if weather is not None:
            if isinstance(weather, str):
                self.world.set_weather(getattr(carla.WeatherParameters, weather))
            elif isinstance(weather, dict):
                self.world.set_weather(carla.WeatherParameters(**weather))

        # Setup HUD
        if self.render:
            self.displayDim = (1280, 720)
            self.displayClock = pygame.time.Clock()
            self.camTransform = 0
            pygame.init()
            pygame.font.init()
            self.hud = visuals.HUD(*self.displayDim)
            self.display = pygame.display.set_mode(
                self.displayDim, pygame.HWSURFACE | pygame.DOUBLEBUF
            )
            self.cameraManager = None

        if self.record:
            if not os.path.exists(self.record):
                os.mkdir(self.record)
            name = "{}/scenario{}.log".format(self.record, self.scenario_number)
            # Carla is looking for an absolute path, so convert it if necessary.
            name = os.path.abspath(name)
            self.client.start_recorder(name)

        # Create objects.
        super().setup()

        # Set up camera manager and collision sensor for ego
        if self.render:
            camIndex = 0
            camPosIndex = 0
            egoActor = self.objects[0].carlaActor
            self.cameraManager = visuals.CameraManager(self.world, egoActor, self.hud)
            self.cameraManager._transform_index = camPosIndex
            self.cameraManager.set_sensor(camIndex)
            self.cameraManager.set_transform(self.camTransform)

        self.world.tick()  ## allowing manualgearshift to take effect    # TODO still need this?

        for obj in self.objects:
            if isinstance(obj.carlaActor, carla.Vehicle):
                obj.carlaActor.apply_control(
                    carla.VehicleControl(manual_gear_shift=False)
                )

        self.world.tick()

        for obj in self.objects:
            if obj.speed is not None and obj.speed != 0:
                raise RuntimeError(
                    f"object {obj} cannot have a nonzero initial speed "
                    "(this is not yet possible in CARLA)"
                )

    def createObjectInSimulator(self, obj):
        # Extract blueprint
        try:
            blueprint = self.blueprintLib.find(obj.blueprint)
        except IndexError as e:
            found = False
            if obj.blueprint in oldBlueprintNames:
                for oldName in oldBlueprintNames[obj.blueprint]:
                    try:
                        blueprint = self.blueprintLib.find(oldName)
                        found = True
                        warnings.warn(
                            f"CARLA blueprint {obj.blueprint} not found; "
                            f"using older version {oldName}"
                        )
                        obj.blueprint = oldName
                        break
                    except IndexError:
                        continue
            if not found:
                raise SimulationCreationError(
                    f"Unable to find blueprint {obj.blueprint}" f" for object {obj}"
                ) from e
        if obj.rolename is not None:
            blueprint.set_attribute("role_name", obj.rolename)

        # set walker as not invincible
        if blueprint.has_attribute("is_invincible"):
            blueprint.set_attribute("is_invincible", "False")

        # Set up transform
        loc = utils.scenicToCarlaLocation(
            obj.position,
            world=self.world,
            blueprint=obj.blueprint,
            snapToGround=obj.snapToGround,
        )
        rot = utils.scenicToCarlaRotation(obj.orientation)
        transform = carla.Transform(loc, rot)

        # Color, cannot be set for Pedestrians
        if blueprint.has_attribute("color") and obj.color is not None:
            c = obj.color
            c_str = f"{int(c.r*255)},{int(c.g*255)},{int(c.b*255)}"
            blueprint.set_attribute("color", c_str)

        # Create Carla actor
        carlaActor = self.world.try_spawn_actor(blueprint, transform)
        if carlaActor is None:
            raise SimulationCreationError(f"Unable to spawn object {obj}")
        obj.carlaActor = carlaActor

        carlaActor.set_simulate_physics(obj.physics)

        if isinstance(carlaActor, carla.Vehicle):
            # TODO should get dimensions at compile time, not simulation time
            extent = carlaActor.bounding_box.extent
            ex, ey, ez = extent.x, extent.y, extent.z
            # Ensure each extent is positive to work around CARLA issue #5841
            obj.width = ey * 2 if ey > 0 else obj.width
            obj.length = ex * 2 if ex > 0 else obj.length
            obj.height = ez * 2 if ez > 0 else obj.height
            carlaActor.apply_control(carla.VehicleControl(manual_gear_shift=True, gear=1))
        elif isinstance(carlaActor, carla.Walker):
            carlaActor.apply_control(carla.WalkerControl())
            # spawn walker controller
            controller_bp = self.blueprintLib.find("controller.ai.walker")
            controller = self.world.try_spawn_actor(
                controller_bp, carla.Transform(), carlaActor
            )
            if controller is None:
                raise SimulationCreationError(
                    f"Unable to spawn carla controller for object {obj}"
                )
            obj.carlaController = controller
        return carlaActor

    def executeActions(self, allActions):
        super().executeActions(allActions)

        # Apply control updates which were accumulated while executing the actions
        for obj in self.agents:
            ctrl = obj._control
            if ctrl is not None:
                obj.carlaActor.apply_control(ctrl)
                obj._control = None

    def step(self):
        # Run simulation for one timestep
        self.world.tick()

        # Render simulation
        if self.render:
            self.cameraManager.render(self.display)
            pygame.display.flip()

    def getProperties(self, obj, properties):
        # Extract Carla properties
        carlaActor = obj.carlaActor
        currTransform = carlaActor.get_transform()
        currLoc = currTransform.location
        currRot = currTransform.rotation
        currVel = carlaActor.get_velocity()
        currAngVel = carlaActor.get_angular_velocity()

        # Prepare Scenic object properties
        position = utils.carlaToScenicPosition(currLoc)
        velocity = utils.carlaToScenicPosition(currVel)
        speed = math.hypot(*velocity)
        angularSpeed = utils.carlaToScenicAngularSpeed(currAngVel)
        angularVelocity = utils.carlaToScenicAngularVel(currAngVel)
        globalOrientation = utils.carlaToScenicOrientation(currRot)
        yaw, pitch, roll = obj.parentOrientation.localAnglesFor(globalOrientation)
        elevation = utils.carlaToScenicElevation(currLoc)

        values = dict(
            position=position,
            velocity=velocity,
            speed=speed,
            angularSpeed=angularSpeed,
            angularVelocity=angularVelocity,
            yaw=yaw,
            pitch=pitch,
            roll=roll,
            elevation=elevation,
        )
        return values

    def destroy(self):
        for obj in self.objects:
            if obj.carlaActor is not None:
                if isinstance(obj.carlaActor, carla.Vehicle):
                    obj.carlaActor.set_autopilot(False, self.tm.get_port())
                if isinstance(obj.carlaActor, carla.Walker):
                    obj.carlaController.stop()
                    obj.carlaController.destroy()
                obj.carlaActor.destroy()
        if self.render and self.cameraManager:
            self.cameraManager.destroy_sensor()

        self.client.stop_recorder()

        self.world.tick()
        super().destroy()



import math

import carla
import scipy

from scenic.core.geometry import normalizeAngle
from scenic.core.vectors import Orientation, Vector


def _snapToGround(world, location, blueprint):
    """Mutates @location to have the same z-coordinate as the nearest waypoint in @world."""
    waypoint = world.get_map().get_waypoint(location)
    # patch to avoid the spawn error issue with vehicles and walkers.
    z_offset = 0
    if blueprint is not None and ("vehicle" in blueprint or "walker" in blueprint):
        z_offset = 0.5

    location.z = waypoint.transform.location.z + z_offset
    return location


def scenicToCarlaVector3D(x, y, z=0.0):
    # NOTE: Used for velocity, acceleration; superclass of carla.Location
    z = 0.0 if z is None else z
    return carla.Vector3D(x, -y, z)


def scenicToCarlaLocation(pos, world=None, blueprint=None, snapToGround=False):
    if snapToGround:
        assert world is not None
        return _snapToGround(world, carla.Location(pos.x, -pos.y, 0.0), blueprint)
    return carla.Location(pos.x, -pos.y, pos.z)


def scenicToCarlaRotation(orientation):
    # CARLA uses intrinsic yaw, pitch, roll rotations (in that order), like Scenic,
    # but with yaw being left-handed and with zero yaw being East.
    yaw, pitch, roll = orientation.r.as_euler("ZXY", degrees=True)
    yaw = -yaw - 90
    return carla.Rotation(pitch=pitch, yaw=yaw, roll=roll)


def carlaToScenicPosition(loc):
    return Vector(loc.x, -loc.y, loc.z)


def carlaToScenicElevation(loc):
    return loc.z


def carlaToScenicOrientation(rot):
    angles = (-rot.yaw - 90, rot.pitch, rot.roll)
    r = scipy.spatial.transform.Rotation.from_euler(
        seq="ZXY", angles=angles, degrees=True
    )
    return Orientation(r)


def carlaToScenicHeading(rot):
    return normalizeAngle(-math.radians(rot.yaw + 90))


def carlaToScenicAngularSpeed(vel):
    return math.hypot(math.radians(vel.x), -math.radians(vel.y), math.radians(vel.y))


def carlaToScenicAngularVel(vel):
    return Vector(math.radians(vel.x), -math.radians(vel.y), math.radians(vel.y))


_scenicToCarlaMap = {
    "red": carla.TrafficLightState.Red,
    "green": carla.TrafficLightState.Green,
    "yellow": carla.TrafficLightState.Yellow,
    "off": carla.TrafficLightState.Off,
    "unknown": carla.TrafficLightState.Unknown,
}


def scenicToCarlaTrafficLightStatus(status):
    return _scenicToCarlaMap.get(status, None)


def carlaToScenicTrafficLightStatus(status):
    return str(status).lower()



"""Support for lazy evaluation of expressions and specifiers.

Lazy evaluation is necessary for expressions like :scenic:`30 deg relative to roadDirection`
where :scenic:`roadDirection` is a vector field and so defines a different heading at
different positions. Scenic defers evaluation of such expressions until they are used in
the definition of an object, when the required context (here, a position) is available.
This is implemented by representing lazy values as special objects which capture all
operations applied to them (in a similar way to `Distribution` objects). The main class
of such objects is `DelayedArgument`: in the above example, the :scenic:`relative to`
operator returns such an object. However, since lazy values can appear as arguments to
distributions, `Distribution` objects can also require lazy evaluation (prior to
sampling); therefore both of these classes derive from a common abstract class
`LazilyEvaluable`.
"""

import inspect
import itertools
import types

from scenic.core.utils import DefaultIdentityDict


class LazilyEvaluable:
    """Values which may require evaluation in the context of an object being constructed.

    If a LazilyEvaluable specifies any properties it depends on, then it cannot be evaluated to a
    normal value except during the construction of an object which already has values for those
    properties.

    Args:
        requiredProps: sequence of strings naming all properties which this value can
            depend on (formally, which must exist in the object passed as the context to
            `evaluateIn`).
        dependencies: for internal use only (see `Samplable`).

    Attributes:
        _requiredProperties: set of strings as above.
    """

    def __init__(self, requiredProps, dependencies=()):
        self._dependencies = tuple(dependencies)  # fixed order for reproducibility
        self._requiredProperties = tuple(sorted(set(requiredProps)))
        self._needsSampling = ns = bool(self._dependencies)
        self._needsLazyEval = nle = bool(self._requiredProperties)
        self._isLazy = ns or nle

    def evaluateIn(self, context):
        """Evaluate this value in the context of an object being constructed.

        The object must define all of the properties on which this value depends.
        """
        # cache of lazy values already evaluated in this context
        cache = context._evaluated
        if self in cache:
            return cache[self]  # avoid making a new evaluated copy of this value
        assert all(hasattr(context, prop) for prop in self._requiredProperties)
        value = self.evaluateInner(context)
        # value should not require further evaluation
        assert not needsLazyEvaluation(value)
        cache[self] = value
        return value

    def evaluateInner(self, context):
        """Actually evaluate in the given context, which provides all required properties.

        Overridden by subclasses.
        """
        return self

    @staticmethod
    def makeContext(**props):
        """Make a context with the given properties."""
        context = types.SimpleNamespace(**props)
        context._evaluated = DefaultIdentityDict()
        return context

    @staticmethod
    def getContextValues(context):
        properties = context.__dict__.copy()
        del properties["_evaluated"]
        return properties


class DelayedArgument(LazilyEvaluable):
    """DelayedArgument(requiredProps, value, _internal=False)

    Specifier arguments requiring other properties to be evaluated first.

    The value of a DelayedArgument is given by a function mapping the context (object under
    construction) to a value.

    .. note::

        When called from a dynamic behavior, constructors for delayed arguments return
        *actual evaluations*, not `DelayedArgument` objects. The agent running the
        behavior is used as the evaluation context.

    Args:
        requiredProps: see `LazilyEvaluable`.
        value: function taking a single argument (the context) and returning the
            corresponding evaluation of this object.
        _internal (bool): set to `True` for internal uses that need to suppress the
            exceptional handling of calls from dynamic behaviors above.
    """

    def __new__(cls, *args, _internal=False, **kwargs):
        darg = super().__new__(cls)
        if _internal:
            return darg
        # at runtime, evaluate immediately in the context of the current agent
        import scenic.syntax.veneer as veneer

        if veneer.simulationInProgress() and veneer.currentBehavior:
            agent = veneer.currentBehavior._agent
            assert agent
            darg.__init__(*args, **kwargs)
            agent._evaluated = DefaultIdentityDict()
            value = darg.evaluateIn(agent)
            del agent._evaluated
            return value
        else:
            return darg

    def __init__(self, requiredProps, value, _internal=False):
        self.value = value
        super().__init__(requiredProps)
        self._needsLazyEval = True
        self._isLazy = True

    def evaluateInner(self, context):
        return self.value(context)

    def __getattr__(self, name):
        if name.startswith("__") and name.endswith("__"):  # ignore special attributes
            return object.__getattribute__(self, name)
        return DelayedArgument(
            self._requiredProperties,
            lambda context: getattr(self.evaluateIn(context), name),
            _internal=True,
        )

    def __call__(self, *args, **kwargs):
        subprops = (
            requiredProperties(arg) for arg in itertools.chain(args, kwargs.values())
        )
        props = set(self._requiredProperties).union(*subprops)

        def value(context):
            subvalues = (valueInContext(arg, context) for arg in args)
            kwsvs = {name: valueInContext(arg, context) for name, arg in kwargs.items()}
            return self.evaluateIn(context)(*subvalues, **kwsvs)

        return DelayedArgument(props, value, _internal=True)


# Operators which can be applied to DelayedArguments
allowedOperators = [
    "__neg__",
    "__pos__",
    "__abs__",
    "__lt__",
    "__le__",
    "__eq__",
    "__ne__",
    "__gt__",
    "__ge__",
    "__add__",
    "__radd__",
    "__sub__",
    "__rsub__",
    "__mul__",
    "__rmul__",
    "__truediv__",
    "__rtruediv__",
    "__floordiv__",
    "__rfloordiv__",
    "__mod__",
    "__rmod__",
    "__divmod__",
    "__rdivmod__",
    "__pow__",
    "__rpow__",
    "__round__",
    "__len__",
    "__getitem__",
]


def makeDelayedOperatorHandler(op):
    def handler(self, *args):
        props = set(self._requiredProperties).union(
            *(requiredProperties(arg) for arg in args)
        )

        def value(context):
            subvalues = (valueInContext(arg, context) for arg in args)
            return getattr(self.evaluateIn(context), op)(*subvalues)

        return DelayedArgument(props, value, _internal=True)

    return handler


for op in allowedOperators:
    setattr(DelayedArgument, op, makeDelayedOperatorHandler(op))


def makeDelayedFunctionCall(func, args, kwargs={}):
    """Utility function for creating a lazily-evaluated function call."""
    assert callable(func), func
    props = set().union(
        *(requiredProperties(arg) for arg in itertools.chain(args, kwargs.values()))
    )

    def value(context):
        subvalues = (valueInContext(arg, context) for arg in args)
        kwsubvals = {name: valueInContext(arg, context) for name, arg in kwargs.items()}
        return func(*subvalues, **kwsubvals)

    return DelayedArgument(props, value, _internal=True)


def valueInContext(value, context):
    """Evaluate something in the context of an object being constructed."""
    if isinstance(value, LazilyEvaluable) and needsLazyEvaluation(value):
        return value.evaluateIn(context)
    return value


def toLazyValue(thing):
    """Wrap a Python object in a `DelayedArgument` if it needs lazy evaluation."""
    if isinstance(thing, DelayedArgument):
        return thing
    if isinstance(thing, (list, tuple)):
        coords = tuple(toLazyValue(c) for c in thing)
        if any(needsLazyEvaluation(c) for c in coords):
            if isinstance(thing, tuple) and hasattr(thing, "_fields"):  # namedtuple
                builder = type(thing)._make
            else:
                builder = type(thing)
            packer = lambda *args: builder(args)
            return makeDelayedFunctionCall(packer, coords)
    elif isinstance(thing, dict):
        items = toLazyValue(tuple(thing.items()))
        if needsLazyEvaluation(items):
            return makeDelayedFunctionCall(type(thing), (items,))
    return thing


def requiredProperties(thing):
    """Set of properties needed to evaluate the given value, if any."""
    if isinstance(thing, LazilyEvaluable):
        return set(thing._requiredProperties)
    else:
        return set()


def needsLazyEvaluation(thing):
    """Whether the given value requires lazy evaluation."""
    return getattr(thing, "_needsLazyEval", False)


def dependencies(thing):
    """Dependencies which must be sampled before this value."""
    return getattr(getattr(thing, "_conditioned", thing), "_dependencies", ())


def needsSampling(thing):
    """Whether this value requires sampling."""
    return getattr(thing, "_needsSampling", False)


def isLazy(thing):
    """Whether this value requires either sampling or lazy evaluation."""
    return getattr(thing, "_isLazy", False)
